import os


class FileNotFound(IOError):
    """
    Raised if any of the files required by the the Docker Build tool is not found
    """
    pass


class FileLoader(object):
    """
    Loads the content of a file

    :param path: The path of the file that is to be loaded

    :type path: str

    :raises FileNotFound: If no file is found for the given path
    """
    def __init__(self, path):

        if not path:
            raise ValueError("File path must be specified and cannot be None")

        self._path = os.path.expanduser(path)
        self._content = None

    @property
    def content(self):
        return self._content

    def exists(self):
        return os.path.exists(self._path)

    def load(self):
        """
        Loads the file in memory, to be processed

        :return: The parser itself
        :raises FileNotFound: If the file is not found in the given path

        :rtype: FileLoader
        """
        # load the build file
        self._content = self._read()
        return self

    def _read(self):
        """
        Reads the file and loads it into the memory for parsing

        :return: The content of the build file

        :rtype: str
        """
        # determine if the file exists
        if not self.exists():
            raise FileNotFound(
                "File not found at {!r}, please make sure that the right "
                "path was specified".format(self._path)
            )

        with open(self._path) as build_file:
            return build_file.read()



import re

from string import Formatter
from exception import \
    InvalidArgumentReference, \
    InvalidFunctionReference, \
    FunctionExecutionError


# the list of functions that can be used in the configuration
Functions = {
    "lower": lambda value: str(value).lower(),
    "upper": lambda value: str(value).upper(),
    "capitalise": lambda value: str(value).capitalize()
}


class ConfigurationParser():

    def __init__(self):
        self._parser = Formatter()

    @staticmethod
    def parse(configurations, build_arguments=None):
        """
        Parses the given configuration to evaluate all the found arguments and functions. All the
        found arguments will be replaced with the value of that argument. All functions will be
        evaluated and the result to the evaluation will be used to replace the referenced function

        :param configurations: The configuration that is to be parsed
        :param build_arguments: The list of known build arguments

        :return: The parsed configuration with all the arguments and functions evaluated

        :type configurations: str
        :type build_arguments: dict

        :rtype: str
        """
        if not configurations:
            raise ValueError("Configurations must be specified and cannot be None")

        parser = Formatter()
        arguments = build_arguments or {}
        parsed_configuration = []

        # start formatting the configuration. For the scope of the first version of the parser the
        # format
        for literal_text, expression, format_spec, conv_spec in parser.parse(configurations):

            # no need to do anything to the literal text, copy as is
            parsed_configuration.append(literal_text)

            # if any expression is specified try to evaluate it
            if expression:
                # append to the resulting configuration to be concatenated at the end
                parsed_configuration.append(
                    ConfigurationParser._evaluate_expression(
                        parser, expression, arguments, format_spec, conv_spec
                    )
                )

        # concatenate the parsed configuration and return, all done
        return "".join(parsed_configuration)

    @staticmethod
    def _evaluate_expression(parser, expression, arguments, format_spec=None, conv_spec=None):
        """
        Evaluates the expression that was given. The expression can be either an argument or a
        function. The expression is evaluated and the value of which returned.

        :param expression: The expression that is to be evaluated
        :param arguments: The list of arguments that is being used for the build process
        :param format_spec: The format spec that should be used on the result of the evaluated
            expression. Optional
        :param conv_spec: The conversion spec that should be used on the result of the evaluated
            expression. Optional

        :return: The result of the expression

        :type expression: str
        :type arguments: dict
        :type format_spec: str
        :type conv_spec: str

        :rtype: str
        """

        function_details = re.match("^([a-zA-Z0-9_-]+)\((.+)\)", expression)

        if function_details:
            # execute the function
            value = ConfigurationParser._execute_function(
                name=function_details.groups()[0],
                parameters=function_details.groups()[1],
                arguments=arguments
            )
        else:
            try:
                # parse the argument
                value, not_req = parser.get_field(
                    field_name=expression,
                    args=None,
                    kwargs=arguments
                )
            except KeyError:
                raise InvalidArgumentReference(
                    "Referenced argument {!r} does not exist, please make sure that the "
                    "correct spelling is used or declare the argument".format(expression)
                )

        # format the value if formatting was specified
        if format_spec:
            value = parser.format_field(value, format_spec)

        # convert the value if a conversion spec was specified
        if conv_spec:
            value = parser.convert_field(value, conv_spec)

        return value

    @staticmethod
    def _execute_function(name, parameters, arguments):
        """
        Executes the function and returns the value for the evaluation

        :param name: The name of the function to be executed
        :param parameters: The parameters as a comma separated list
        :param arguments: The list of arguments that is being used for the build process

        :return: The result of executing the function

        :type name: str
        :type parameters: str
        :type arguments: dict

        :rtype: str

        :raises InvalidFunctionReference: Raised if the given name of the function is not known the
            the build tool
        """

        # confirm if the function exists
        if name not in Functions:
            raise InvalidFunctionReference(
                "Referenced function {!r} is not valid, please make sure that the "
                "correct spelling is used".format(name)
            )

        # parse the parameters and evaluate any arguments or functions
        parameter_list = ConfigurationParser.parse(parameters, arguments).split(",")

        try:

            # run the function
            return Functions[name](*parameter_list)

        except TypeError as ex:
            raise FunctionExecutionError(
                "Execution of function {!r} failed due to error: {!s}".format(name, ex)
            )



"""
Defines the different enumerations that are required by the Docker Build tool
"""

import types

from enum import Enum
from docker_build.utils.argparser import convert_to_list


class Configuration(Enum):
    """
    Details of a configuration used to set a Docker image
    """
    CMD = ("Cmd", [types.ListType, types.StringType, types.NoneType])
    ENTRYPOINT = ("Entrypoint", [types.ListType, types.StringType, types.NoneType])
    ENV = ("Env", [types.DictType], convert_to_list)
    EXPOSE = ("ExposedPorts", [types.DictType])
    LABELS = ("Labels", [types.DictType])
    ONBUILD = ("OnBuild", [types.ListType])
    USER = ("User", [types.StringType])
    VOLUMES = ("Volumes", [types.ListType])
    WORKDIR = ("WorkingDir", [types.StringType])
    STOPSIGNAL = ("StopSignal", [types.StringType])

    def __init__(self, docker_command, supported_types, conversion_fn=None):
        self.docker_command = docker_command
        self.supported_types = supported_types
        self.conversion_fn = conversion_fn

    def validate_value(self, value):
        if type(value) not in self.supported_types:
            raise TypeError(
                "Configuration {!r} value is not valid, type should be one of {!r}".format(
                    self.name,
                    self.supported_types
                )
            )

    def convert_value(self, value):
        return self.conversion_fn(value) if self.conversion_fn else value



import copy

import yaml
from docker_build.constants import BUILD_CONTEXT_DST_PATH
from docker_build.configuration.encoder import decode_argument_value
from docker_build.configuration.exception import \
    InvalidMainConfigurations, \
    InvalidBuildConfigurations, \
    MissingArgument, \
    InvalidArgumentValue, \
    InvalidArgumentMapping
from docker_build.configuration.parser import ConfigurationParser
from yaml.parser import ParserError


class MainConfig(object):
    """
    The main configurations for a Docker build
    """

    def __init__(self, config=None):
        self._config = config

        self._parsed_config = self._parse() if config else None
        self._arguments = self._read_arguments()

    @property
    def arguments(self):
        """
        The list of arguments that have been loaded from the configurations

        :rtype: dict
        """
        return self._arguments

    def _parse(self):

        try:

            return yaml.load(self._config)

        except ParserError as ex:
            raise InvalidMainConfigurations(
                "Main configurations are invalid. Configurations parsing failed with "
                "error {!r} at {!r}"
                .format(
                    ex.problem,
                    str(ex.problem_mark)
                )
            )

    def _read_arguments(self):
        """
        Reads the arguments from the build tool configurations. These are common configurations that
        are used across all builds

        :return: The list of arguments that have been found in the configurations
        :rtype: dict
        """
        arguments = {}

        if self._config and "ARGS" in self._config:

            try:

                # read all the arguments
                for name, attributes in self._parsed_config["ARGS"].items():
                    try:

                        value = attributes["VALUE"]
                        is_encrypted = "ENCRYPTED" in attributes and attributes["ENCRYPTED"]
                        arguments[name] = \
                            decode_argument_value(name, value) if is_encrypted else value

                    except KeyError as ex:
                        raise InvalidMainConfigurations(
                            "Argument {!r} is not properly configured, attribute {!s} is missing"
                            .format(name, ex)
                        )

            except Exception as ex:
                raise InvalidMainConfigurations(
                    "Main configurations contains invalid argument declaration, parsing of "
                    "configurations failed with error - {!s}".format(ex)
                )

        return arguments


class BuildConfig(object):
    """
    The build config containing instructions of a particular build that is to be performed

    :param build_details: The details of a build as loaded from the source
    :param build_arguments: The list of arguments as specified for the build

    :type build_details: str
    :type build_arguments: dict
    """

    def __init__(self, build_details, build_arguments=None):

        if not build_details:
            raise ValueError("Build details must be specified and cannot be None")

        self._build_details = build_details

        # parse the build file and read the arguments
        self._parsed_build_details = self._parse()
        self._build_arguments = self._read_arguments(build_arguments or {})

        # reload the build file populating the build arguments
        self._parsed_build_details = self._parse()

    @property
    def build_details(self):
        return self._parsed_build_details

    def _parse(self):

        build_details = copy.copy(self._build_details)

        try:

            if hasattr(self, "_build_arguments") and self._build_arguments:
                build_details = ConfigurationParser.parse(build_details, self._build_arguments)

            return yaml.load(build_details)

        except KeyError as ex:
            raise InvalidBuildConfigurations(
                "Build configurations are invalid. Argument {!r} is not defined".format(
                    ex.message
                )
            )

        except ParserError as ex:
            raise InvalidBuildConfigurations(
                "Build configurations are invalid. Details failed with error {!r} at {!r}".format(
                    ex.problem,
                    str(ex.problem_mark)
                )
            )

    def _read_arguments(self, build_arguments):

        # the list of variables that are loaded from the list of arguments for the build
        arguments = copy.deepcopy(build_arguments)

        if "ARGS" in self._parsed_build_details:

            try:

                # read all the arguments
                for name, attributes in self._parsed_build_details["ARGS"].items():

                    # if an argument is set as required confirm that the value for the argument is
                    # known. if on the other hand the argument is optional confirm that a default
                    # was given
                    if "REQUIRED" in attributes and attributes["REQUIRED"]:
                        if name not in arguments:
                            raise MissingArgument(
                                "Build argument {!r} is required but no value was passed in for "
                                "the argument".format(name)
                            )
                    else:
                        if "DEFAULT" not in attributes:
                            raise MissingArgument(
                                "Build argument {!r} is required but no default value is specified"
                                .format(name)
                            )

                    # populate the default for the argument if it was not passed
                    if "DEFAULT" in attributes and name not in arguments:
                        arguments[name] = attributes["DEFAULT"]

                    # confirm that the right value was given for the argument
                    if "CHOICES" in attributes and name in arguments:
                        if arguments[name] not in attributes["CHOICES"]:
                            raise InvalidArgumentValue(
                                "Value {value!r} for build argument {name!r} is invalid, supported "
                                "values are {choices!r}".format(
                                    value=arguments[name],
                                    name=name,
                                    choices=attributes["CHOICES"]
                                )
                            )

                    # confirm if there are any other variables to be loaded
                    if "MAPPINGS" in attributes:

                        for index, mapping in enumerate(attributes["MAPPINGS"]):

                            if "NAME" not in mapping:
                                raise InvalidArgumentMapping(
                                    "Mapping [{mapping_index}] for build argument {argument_name!r}"
                                    " is invalid, mapping should contain NAME attribute".format(
                                        mapping_index=index,
                                        argument_name=name
                                    )
                                )

                            mapping_name = mapping["NAME"]

                            if "VALUES" not in mapping:
                                raise InvalidArgumentMapping(
                                    "Mapping {mapping_name!r} for build argument {argument_name!r} "
                                    "is invalid, mapping should contain VALUES attribute".format(
                                        mapping_name=mapping_name,
                                        argument_name=name
                                    )
                                )

                            argument_value = arguments[name]
                            mapping_values = mapping["VALUES"]
                            mapping_default = mapping["DEFAULT"] if "DEFAULT" in mapping else None

                            if argument_value not in mapping_values and mapping_default is None:
                                raise InvalidArgumentMapping(
                                    "Mapping {mapping_name!r} for argument {argument_name!r} does "
                                    "not contain mapping for value {value!r} and no default value "
                                    "specified either".format(
                                        mapping_name=mapping_name,
                                        argument_name=name,
                                        value=argument_value
                                    )
                                )

                            # add the new variable to the list of build arguments
                            arguments[mapping_name] = mapping_values[argument_value] \
                                if argument_value in mapping_values else mapping_default

                            # confirm that no data structures have been passed and only string or
                            # numbers have been set for the value
                            if isinstance(arguments[mapping_name], dict) or \
                                    isinstance(arguments[mapping_name], list):
                                raise InvalidArgumentMapping(
                                    "Mapping {mapping_name!r} for argument {argument_name!r} "
                                    "contains an invalid mapping for value {value!r}, only strings "
                                    "and numbers are supported".format(
                                        mapping_name=mapping_name,
                                        argument_name=name,
                                        value=argument_value
                                    )
                                )

            except Exception as ex:
                raise InvalidBuildConfigurations(
                    "Build configurations contains invalid argument declaration, parsing of "
                    "details failed with error - {!s}".format(
                        ex
                    )
                )

        # inject the build context path (path inside the container) that can be used for
        # reference during the build process
        arguments["BUILD_CONTEXT_PATH"] = BUILD_CONTEXT_DST_PATH

        return arguments



import os

from docker_build.utils.loader import FileLoader

# the default path for the main configuration file
MAIN_CONFIG_FILE_PATH = "~/.docker/build-config.yml"


class MainConfigFileLoader(FileLoader):

    def __init__(self, path=None):
        super(MainConfigFileLoader, self).__init__(
            path=MAIN_CONFIG_FILE_PATH if not path else path
        )

    def load(self):

        # if the main configuration is to be loaded from the default path and no file exists it
        # means that the user did not set any main configurations which is perfectly valid. In that
        # case no main configurations should be loaded.
        if self._path == os.path.expanduser(MAIN_CONFIG_FILE_PATH) and not self.exists():
            return

        return FileLoader.load(self)



"""
Build tool to create Docker images. This is another implementation of the builder different than the
one provided by the official repository. This tool is intended to give the user more flexibility for
the creation of an image as opposed to what is being provided by the build command provided in
docker
"""
from __future__ import print_function

import argparse
import json
import logging
import io
import tarfile
import sys
import docker
import types
import os
import time

from docker import errors
from docker_build import __version__
from docker_build import exception
from docker_build.configuration.loader import FileLoader, MainConfigFileLoader
from docker_build.configuration.model import BuildConfig, MainConfig
from docker_build.constants import BUILD_CONTEXT_DST_PATH
from docker_build.daemon.catalog import Configuration
from docker_build.utils.argparser import \
    PutAction, \
    parse_key_value_option
from docker_build.utils.logger import ConsoleLogger
from requests.exceptions import RequestException


# the logger for the docker build tool
log = logging.getLogger("docker_build")


def _parse_config(configs, parsed_configs, configuration_option):
    """
    Validates the given configuration and if required converts the configuration from the format
    supported by the Docker Build tool to the one understood by Docker Daemon.
    """
    if configuration_option.name in configs:

        # get the value of the passed in configuration option
        value = configs[configuration_option.name]

        # validate the configuration value
        configuration_option.validate_value(value)

        # convert the value to the one supported by Docker Daemon
        parsed_configs[configuration_option.docker_command] = configuration_option.convert_value(
            value
        )


def _inspect_image(docker_client, image):
    """
    Inspect the details of the image returning back the full details of that image
    """
    details = None

    try:
        details = docker_client.inspect_image(image)
    except errors.NotFound:
        pass

    return details


def _inspect_container(docker_client, container):
    """
    Inspect the details of the container returning back the full details of that container
    """
    details = None

    try:
        details = docker_client.inspect_container(container)
    except errors.NotFound:
        pass

    return details


def _pull_image(docker_client, image_name):
    """
    Pulls the Docker Image from the remote Docker Registry
    """
    status_log = None
    refresh_count = 1
    repository, tag = _get_docker_image_name_parts(image_name)
    params = {
        "repository": repository,
        "tag": tag,
        "stream": True
    }

    for output in docker_client.pull(**params):
        details = json.loads(output)

        if "status" in details:
            if not status_log:
                log.info(details["status"])
            else:
                print("#" * refresh_count, end="\r")
                refresh_count = refresh_count + 1 if refresh_count <= 50 else 1

            status_log = details["status"]

        if "error" in details:
            raise exception.DockerImageNotFound(details["error"])

    print("", end="\r")
    log.info(status_log)


def _create_container(docker_client, image):
    """
    Create a container that will be used to execute the commands and create the new required image.
    The image will be created and started.
    """

    # the list of parameters that will be passed to the docker command
    params = {
        "tty": True,
        "detach": True,
        "command": "/bin/sh",
        "image": image
    }

    # if the image that the container is being started from has an entry point overwrite it to clear
    # the entry point
    details = _inspect_image(docker_client, image)
    if details and details["Config"]["Entrypoint"]:
        params["entrypoint"] = []

    def create_container_with_auto_pull(remote_download_tried=False):
        # create the container that will be used to run the details for the image
        try:
            return docker_client.create_container(**params)
        except errors.NotFound:
            if not remote_download_tried:
                log.info("Image {!r} not found locally, trying remote registry".format(image))
                _pull_image(docker_client, image)
                return create_container_with_auto_pull(True)
            else:
                raise exception.DockerImageNotFound("Image {!r} could not be found".format(image))

    # create the container
    container = create_container_with_auto_pull()

    if "Warnings" in container and container["Warnings"]:
        log.warn("Created container contains warnings {!r}".format(container["Warnings"]))

    container_id = container["Id"]

    # start the container
    docker_client.start(container=container_id)

    return container_id


def _copy(docker_client, container_id, source, destination):
    """
    Copies a file or directory from a given local path to the container being used for the build
    """

    log.debug("Copying content from {!r} to container path {!r}".format(source, destination))

    # confirm that the given path is valid
    if not os.path.exists(source):
        raise exception.SourcePathNotFound(
            "Source path {!r} is invalid, specified path could not be found".format(source)
        )

    # determine the source and destination type
    is_src_dir = os.path.isdir(source)
    is_dst_dir = destination[-1:] == "/"

    # confirm that the right combination of source to destination is specified
    # the only invalid option is if the source is a directory and the destination is a file
    if is_src_dir and not is_dst_dir:
        raise exception.InvalidDockerBuildOptionValue(
            "Invalid copy destination {!r}, path must be a folder since source {!r} ia a folder"
            .format(destination, source)
        )

    # determine the destination directory according to the determined destination type
    dst_folder = destination if is_dst_dir else os.path.dirname(destination)
    # determine the archive name according to the source and destination
    archive_name = os.path.basename(source) if is_dst_dir else os.path.basename(destination)

    # the in memory archive that will be used to copy the content over to the container
    archive = io.BytesIO()

    # create a tar file with all the contents of the given path
    with tarfile.open(fileobj=archive, mode='w') as tar:
        tar.add(
            name=source,
            arcname=archive_name
        )

    # create the destination folder in the container if it doesn't exist
    _run_command(
        docker_client,
        container_id,
        "mkdir -p {path}".format(path=dst_folder)
    )

    # copy over the content to the container
    docker_client.put_archive(
        container=container_id,
        path=dst_folder,
        data=archive.getvalue()
    )


def _run_command(docker_client, container_id, command, args={}, show_logs=False):
    """
    Runs the given command in the container
    """

    def execute_instructions(instruction_list, variable_list, logger):
        """
        Executes all the given instructions against the container
        """

        # execute the instruction
        execute = docker_client.exec_create(
            container=container_id,
            cmd=[
                "/bin/sh",
                "-c",
                "; ".join(["set -e"] + variable_list + instruction_list)
            ],
            user="root"
        )

        stream = docker_client.exec_start(
            exec_id=execute["Id"],
            stream=True
        )

        # display whatever is being printed to the stdout of the container
        for log_stream in stream:
            logger.log(log_stream)

        # confirm that the command finished with no error
        exit_code = docker_client.exec_inspect(execute["Id"])["ExitCode"]

        if exit_code:
            raise exception.CommandExecutionError(
                "RUN command with instruction/s {instruction!r} failed with exit code [{exit_code}]"
                .format(
                    instruction=instruction_list[0]
                    if len(instruction_list[0]) <= 30 and len(instruction_list) == 1
                    else "{}...".format(instruction_list[0][:30]),
                    exit_code=exit_code
                )
            )

    # the list of variables that will be used during the execution of each command
    environment_variables = [
        "export {name}={value}".format(name=name, value=value)
        for name, value in args.items()
    ]

    # the list of instructions to execute against the container
    instructions = command if isinstance(command, types.ListType) else [command]

    with ConsoleLogger(show_logs, "Start of Container Logs") as console_log:
        execute_instructions(instructions, environment_variables, console_log)


def _commit_image(docker_client, container_id, author=None, configs=None, tag=None):
    """
    Commits the made changes in the container into an image.
    """

    # the list of parameters that will be passed to the docker command
    params = {
        "container": container_id,
        "conf": {}
    }

    # add the tag that should be used for the image to be created if any
    if tag:
        params["repository"], params["tag"] = _get_docker_image_name_parts(tag)

    # populate all other optional parameters
    if author:
        params["author"] = author

    # add all the specified build options
    if configs:
        for configuration_option in Configuration:
            _parse_config(configs, params["conf"], configuration_option)

    # commit the changes
    image = docker_client.commit(**params)
    image_id = image["Id"]

    return str(image_id[7:19])


def _remove_container(docker_client, container_id):
    """
    Removes the container
    """
    # determine if the container is paused first, if it is first un-pause it before trying to remove
    # the container
    if _inspect_container(docker_client, container_id)["State"]["Paused"]:
        docker_client.unpause(container=container_id)

    # remove the container
    docker_client.remove_container(container=container_id, force=True)


def _copy_build_context(docker_client, container_id, step_config):
    """
    Copies the build context to the running container. The build context can be either one or many
    paths that can be copied into the container
    """

    files_copied = False

    if "BUILDCONTEXT" in step_config:

        log.info("Copying building context to the container")
        files_copied = True

        if isinstance(step_config["BUILDCONTEXT"], str):

            _copy(
                docker_client,
                container_id,
                step_config["BUILDCONTEXT"],
                os.path.join(BUILD_CONTEXT_DST_PATH, "")
            )

        elif isinstance(step_config["BUILDCONTEXT"], list):

            for copy_details in step_config["BUILDCONTEXT"]:

                dst = ""

                if "DST" in copy_details:
                    dst = "." + copy_details["DST"] if copy_details["DST"].startswith("/") \
                          else copy_details["DST"]

                dst = os.path.join(BUILD_CONTEXT_DST_PATH, dst)

                if not os.path.normpath(dst).startswith(BUILD_CONTEXT_DST_PATH):
                    raise exception.InvalidDockerBuildFile(
                        "Invalid Build Context 'DST' property {!r}, destination path must be "
                        "within the Build Context folder".format(
                            copy_details["DST"]
                        )
                    )

                _copy(docker_client, container_id, copy_details["SRC"], dst)

        else:

            raise exception.InvalidDockerBuildConfigFile(
                "BUILDCONTEXT is invalid, context must be either a String or a List of SRC and DST "
                "objects"
            )

    return files_copied


def _build(docker_client, args, build_config, step_config, from_image, should_remove_container):
    """
    Builds the image for the given step

    :param docker_client: The Docker Client that is being used to send commands to the Docker Daemon
    :param args: The list of args that are known for the build
    :param build_config: The configurations of the entire build
    :param step_config: The configurations of the step being build with this build process
    :param from_image: The identifier or tag of the image to be used as the base for the image being
                       created
    :param should_remove_container: Indicates if the container should be removed on success or
                                    failure build

    :returns: The identifier of the image that was created

    :type docker_client: docker.Client
    :type args: dict
    :type build_config: dict
    :type step_config: dict
    :type from_image: str
    :type should_remove_container: bool
    :rtype: str
    """
    container_id = None

    try:

        # create the container that will be used to run the details for the image
        log.info("Starting new container from {!r}".format(from_image))
        container_id = _create_container(docker_client, from_image)

        # determine if there is a build context specified
        build_context_populated = _copy_build_context(docker_client, container_id, step_config)

        # copy over any files that are required if any specified
        if "COPY" in step_config:
            log.info("Copying folders or files to container")
            for copy_details in step_config["COPY"]:
                _copy(
                    docker_client,
                    container_id,
                    copy_details["SRC"],
                    copy_details["DST"]
                )

        # execute the commands to make the necessary changes
        if "RUN" in step_config:
            log.info("Making necessary changes to the container")
            _run_command(
                docker_client,
                container_id,
                step_config["RUN"],
                args=args,
                show_logs=True
            )

        # clean up the build context if one was created
        if build_context_populated:
            log.info("Cleaning up container from build context")
            _run_command(
                docker_client,
                container_id,
                "rm -rf {dst}".format(dst=BUILD_CONTEXT_DST_PATH)
            )

        # commit the change done to the container
        log.info("Creating image with container changes")

        # determine if it is the last build step in the process
        is_last_build_step = step_config == build_config["STEPS"][-1]

        # get the configs of the image that was used as the base image
        image_configs = _inspect_image(docker_client, from_image)["Config"]

        # build the configuration that will be set for the image being created
        configs = step_config["CONFIG"] if "CONFIG" in step_config else {}

        # if the command and entry point are not being over written by a specific configuration of
        # the new image being created set the command and / or entry point of the from image. This
        # is being done as the container which was created from (the base images) is overwriting the
        # command and entry point to force the start of shell in the container
        if "CMD" not in configs:
            configs["CMD"] = image_configs["Cmd"] or []

        if "ENTRYPOINT" not in configs:
            configs["ENTRYPOINT"] = image_configs["Entrypoint"]

        image_id = _commit_image(
            docker_client,
            container_id,
            author=build_config["MAINTAINER"] if "MAINTAINER" in build_config else None,
            configs=configs,
            tag=build_config["TAG"] if is_last_build_step and "TAG" in build_config else None
        )

        log.info("Successfully created image {!r}".format(image_id))

        # return the identifier of the image that was created from this build step
        return image_id

    finally:

        # if a container was created remove it to clean up
        if container_id and should_remove_container:
            log.info("Cleaning up container")
            _remove_container(docker_client, container_id)


def _get_docker_image_name_parts(image_name):
    """
    Gets the parts of the image name. The name is split into two parts, the repository and the tag
    """
    image_name_parts = image_name.split(":")
    return (
        image_name_parts[0],
        image_name_parts[1] if len(image_name_parts) > 1 else "latest"
    )


def main(argv=None):
    """
    Main function for invoking the Docker Build tool
    """
    # Parse argument list
    parser = argparse.ArgumentParser(
        description="Build tool for creating Docker images"
    )
    parser.add_argument(
        "-a", "--arg",
        dest="build_args",
        type=parse_key_value_option,
        action=PutAction,
        metavar="NAME=VALUE",
        default={},
        help="Passes the value of an argument that is defined in the build file. The arguments "
             "option can be defined multiple times for each argument that is to be populated with "
             "a value. Passes arguments are expected to be Name/Value pairs separated with an "
             "equals sign"
    )
    parser.add_argument(
        "-b", "--build-config-file",
        dest="build_config_file_path",
        type=str,
        default="./docker-build.yml",
        help="The path to the build configuration file that will be used during the build process. "
             "By default the tool will look for the build configuration file in the current "
             "working directory unless overwritten with the use of this option"
    )
    parser.add_argument(
        "-m", "--main-config-file",
        dest="main_config_file_path",
        type=str,
        help="The path to the main configuration file that will be used during the build process. "
             "By default the tool will look for the main configuration file in the users home "
             "directory under the .docker folder. The default behaviour can be overwritten with "
             "the use of this option"
    )
    parser.add_argument(
        "-t", "--tag",
        dest="tag",
        type=str,
        help="If specified the tag value defined in the build file will be overwritten. The tag "
             "is used to commit the final image that is generated by the build tool. Either the "
             "TAG command or the tag option must be specified"
    )
    parser.add_argument(
        "--keep",
        dest="keep_containers",
        action="store_true",
        help="Keeps the intermediate containers after a build. The default behaviour is to remove "
             "all created containers but sometimes it is useful to leave them for debugging "
             "purposes"
    )
    parser.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Prints more logs during the build process to give more context of what is happening"
             "in during a build of an Docker image"
    )
    parser.add_argument(
        "-v", "--version",
        action="version",
        version="docker-build-tool version {}".format(__version__),
        help="Prints the program's version number and exits"
    )

    try:

        # start the timer to find out at the end how long the build took
        start_time = time.time()

        # parse the command line arguments passed to the tool
        command_line_args = parser.parse_args(argv)

        # determine if lower level logging should be enabled
        if command_line_args.verbose:
            log.setLevel(logging.DEBUG)

        # load the configuration file
        config_file_loader = MainConfigFileLoader(command_line_args.main_config_file_path)
        configs = MainConfig(config_file_loader.load().content)

        # load all the build arguments for the build process
        build_args = dict(
            configs.arguments.items() + command_line_args.build_args.items()
        )

        # load the build file
        build_configs = BuildConfig(
            FileLoader(command_line_args.build_config_file_path).load().content,
            build_args
        )

        docker_client = docker.from_env(assert_hostname=False)

        # determine from which image to start
        if "FROM" not in build_configs.build_details:
            raise exception.InvalidDockerBuildFile(
                "FROM is not optional please confirm the build file"
            )

        from_image = build_configs.build_details["FROM"]

        # if the tag command line argument was specified update the tag that is set in the build
        # file
        if command_line_args.tag:
            build_configs.build_details["TAG"] = command_line_args.tag

        # change the working directory to the path where the build file is located before commencing
        # the build. This will make sure that all the paths in the build file are relative to the
        # build file itself
        os.chdir(os.path.dirname(command_line_args.build_config_file_path))

        # go through the steps to create the necessary images
        for step_config in build_configs.build_details["STEPS"]:
            from_image = _build(
                docker_client,
                build_args,
                build_configs.build_details,
                step_config,
                from_image,
                not command_line_args.keep_containers
            )

        total_build_time = int(time.time() - start_time)
        log.info("Build finished in {} min/s {} sec/s".format(
            int(total_build_time / 60),
            total_build_time % 60
        ))

    except KeyboardInterrupt:
        log.info("Docker Build shutdown by user")
        return 130

    except RequestException:
        log.error("Cannot connect to the Docker daemon. Is the docker daemon running on this host?")
        return 1

    except (exception.DockerBuildException, exception.DockerBuildIOError) as ex:
        log.error("Build failed due to error : {}".format(ex))
        return 1

    except Exception as ex:
        log.exception("Unexpected error during build due to error : {}".format(ex))
        return 1


if __name__ == '__main__':
    sys.exit(main())



import base64

from docker_build.configuration.exception import \
    InvalidArgumentValue


def decode_argument_value(name, value):
    """
    Decodes the value of an argument. The value is assumed to be Base64 encoded

    :param name: The name of the argument that is being decoded
    :param value: The encoded value of the argument. It is assumed to be Base64 encoded

    :return: The decoded value of the argument

    :raises ValueError: If either the name or value arguments are not passed
    :raises InvalidDockerBuildArgumentValue: If the value for the argument is invalid

    :type name: str
    :type value: str

    :rtype: str
    """
    if not name:
        raise ValueError("Argument name must be specified and cannot be None")

    if not value:
        raise ValueError("Argument value must be specified and cannot be None")

    try:

        # base64 decode the value of the argument
        # try to encode the argument value after decoding to make sure that the value is
        # valid
        return unicode(base64.b64decode(value), "utf8")

    except TypeError:
        raise InvalidArgumentValue(
            "Argument {!r} is invalid, argument value is not base64 encoded "
            "but argument is marked as obfuscated".format(name)
        )

    except UnicodeDecodeError:
        raise InvalidArgumentValue(
            "Argument {!r} is invalid, argument value is not a valid base64 string. "
            "Please make sure that the string was properly encoded".format(name)
        )



from logging import StreamHandler, Formatter
import logging

__author__ = "Brian Bason"
__version__ = "0.9.1.1"

# create the default handler
default_handler = StreamHandler()
default_handler.setFormatter(
    fmt=Formatter(
        fmt="%(message)s"
    )
)

# create the handler for the container console
container_console_handler = StreamHandler()
container_console_handler.setFormatter(
    fmt=Formatter(
        fmt="%(message)s",
    )
)

# set the root logger
root_logger = logging.getLogger()
root_logger.addHandler(default_handler)
root_logger.setLevel(logging.WARN)

# set the logger for the builder
docker_build_logger = logging.getLogger("docker_build")
docker_build_logger.setLevel(logging.INFO)

# set the logger for the container console output
container_console_logger = logging.getLogger("container_console")
container_console_logger.setLevel(logging.INFO)
container_console_logger.addHandler(container_console_handler)
container_console_logger.propagate = 0



"""
Configurations that are required by the Docker Build, these include the main configurations which
contains details that are common across multiple builds and the build configurations which contains
details for one particular build cycle.
"""


# the path to the build context on the container. This determines where the specified build context
# folder on the build machine will be copied on the container.
BUILD_CONTEXT_DST_PATH = "/tmp/build-context"

# the default path for the configuration file
CONFIG_FILE_PATH = "~/.docker/build-config.yml"



import os

from docker_build import __version__
from setuptools import setup, find_packages


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

setup(

    name='docker-build-tool',
    version=__version__,

    description='Build tool for creating Docker Images',

    url='https://github.com/brian-bason/docker-build-tool',
    author='Brian Bason',
    author_email='brianbason@gmail.com',
    classifiers=[],

    packages=find_packages(exclude=['test*']),
    include_package_data=True,

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'docker-py>=1.8,<2.0',
        'pyYAML>=3.11,<4.0',
        'enum34>=1.1.6'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'test': [
            'coverage==4.0.1',
            'mock==1.3.0',
            'nose==1.3.7',
            'testfixtures==4.3.3'
        ]
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'docker-build=docker_build.__main__:main'
        ]
    }

)



"""
Utility functions for command line argument parsing
"""

import types
import argparse


def convert_to_list(value):
    """
    Converts the given value into a list. If the value is a dictionary the key and value of the
    dictionary will be parsed into a string with the format key=value. Any other type will be
    converted to its string representation and returned as a list

    :param value: The value that is to be converted

    :return: A list for the given value
    """
    if isinstance(value, types.DictType):
        return [
            "{key}={value}".format(key=k, value=v)
            for k, v in value.items()
        ]
    elif isinstance(value, types.ListType):
        return value
    else:
        return [str(value)]


class PutAction(argparse.Action):
    """
    An Arg Parser action class to put given tuple values into a dictionary. The first item of the
    tuple will be used as the key of the dictionary item being added while the second item will be
    used as the value
    """
    def __call__(self, parser, namespace, values, option_string=None):

        if not isinstance(values, tuple):
            raise argparse.ArgumentParser(
                "Value {} must be of type tuple for {}. Make sure that the type being used to parse"
                "the argument returns a tuple".format(
                    values,
                    self.__class__.__name__
                )
            )

        if len(values) != 2:
            raise argparse.ArgumentParser(
                "Value passed to {} must contain 2 values. Found {} instead".format(
                    self.__class__.__name__,
                    len(values)
                )
            )

        attr = getattr(namespace, self.dest) or {}
        attr[values[0]] = values[1]
        setattr(namespace, self.dest, attr)


def parse_key_value_option(option):
    """
    :param option: The option from the command line that is being parsed

    :return: A tuple containing the name and value of the parsed argument

    :type option: str
    :rtype tuple:
    """
    argument_parts = option.split("=")

    if len(argument_parts) != 2:
        raise argparse.ArgumentTypeError(
            "Option {!r} is not valid, option should be in the format NAME=VALUE".format(
                option
            )
        )

    return argument_parts[0], argument_parts[1]



from docker_build.exception import DockerBuildException


class InvalidMainConfigurations(DockerBuildException):
    """
    Raised if the given main configurations are invalid
    """
    pass


class InvalidBuildConfigurations(DockerBuildException):
    """
    Raised if the given build configurations are invalid
    """
    pass


class MissingArgument(DockerBuildException):
    """
    Raised if an argument is required and is not passed to the build tool
    """
    pass


class InvalidArgumentValue(DockerBuildException):
    """
    Raised if an invalid value was specified for an argument
    """
    pass


class InvalidArgumentMapping(DockerBuildException):
    """
    Raised if an invalid mapping for an argument is defined
    """
    pass


class InvalidArgumentReference(DockerBuildException):
    """
    Raised if an invalid reference is made to an argument in a build file
    """
    pass


class InvalidFunctionReference(DockerBuildException):
    """
    Raised if an invalid reference is made to a function in a build file
    """
    pass


class FunctionExecutionError(DockerBuildException):
    """
    Raised if an error is encountered when a build in function is executed
    """
    pass



import logging


class ConsoleLogger():
    """
    Prints the given log messages to a simulated console. What this means is that the messages are
    logged without any dates or other log information.

    :param is_enabled: Determines if the logger is enabled and if messages should be printed to
                       console
    :param header: A string that will be printed at the start of the console logs. If specified this
                   will also trigger the insertion of a footer message. (OPTIONAL)

    :type is_enabled: bool
    :type header: str
    """
    _header_line_character = "-"
    _header_line_marker = _header_line_character * 26
    _log_line_pre_spacer = "-> "

    def __init__(self, is_enabled=True, header=None):
        self._is_enabled = is_enabled
        self._header = header
        self._log = logging.getLogger("container_console")
        self._buffer = None

    def __enter__(self):
        if self._is_enabled and self._header:
            self._log.info(
                "{line_marker} {header} {line_marker}".format(
                    line_marker=self._header_line_marker,
                    header=self._header
                )
            )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._is_enabled:
            # print whatever is left in the buffer
            if self._buffer:
                self._log.info("{spacer}{message}".format(
                    spacer=self._log_line_pre_spacer,
                    message=self._buffer
                ))

            # print the footer if a header was specified
            if self._header:
                self._log.info(
                    "{line_marker}{header}{line_marker}".format(
                        line_marker=self._header_line_marker,
                        header=self._header_line_character * (len(self._header) + 2)
                    )
                )

    def log(self, message):
        """
        Logs the message to the simulated console

        :param message: The messages that is to be printed

        :type message: str
        """
        if self._is_enabled:

            is_last_message_incomplete = message[-1:] != "\n"

            # split the stream into individual lines, removing any empty lines but first append any
            # previous messages that are in the buffer
            log_lines = "{previous_message}{new_message}".format(
                previous_message=self._buffer if self._buffer else "",
                new_message=message if is_last_message_incomplete else message[:-1]
            ).split("\n")

            # if the last log entry is not complete keep it in the buffer for the next iteration
            # of the log print
            if is_last_message_incomplete:
                self._buffer = log_lines[-1]
                del log_lines[-1]
            else:
                self._buffer = None

            for log_line in log_lines:
                self._log.info("{spacer}{message}".format(
                    spacer=self._log_line_pre_spacer,
                    message=log_line
                ))



"""
The exceptions that can be raised by the build tool
"""


class DockerBuildException(Exception):
    """
    The base class for any application exception raised in the application
    """
    pass


class DockerBuildIOError(IOError):
    """
    The base class for any IO error raised in the application
    """
    pass


class InvalidDockerBuildOptionValue(DockerBuildException):
    """
    Raised if the given Docker build file contains an invalid option value
    """
    pass


class CommandExecutionError(DockerBuildException):
    """
    Raised if the execution of a command in a Docker container failed due to some error
    """
    pass


class DockerImageNotFound(DockerBuildException):
    """
    Raised if the Docker image to be used as the base image for a build could not be found locally
    and on the remote Docker registry
    """
    pass


class SourcePathNotFound(DockerBuildIOError):
    """
    Raised if the given source for a copy operation to a Docker container is not found
    """
    pass


"""
Groups together the functionality that is required to communicate with the Docker daemon that the
build tool will use to perform the build. The package contains the commands that can be sent to the
daemon through the rest APIs and also other functionality required for such communication.
"""